import copy
import numpy
import netCDF4
from operator import mul
from .field import _parse_index, Coordinate
from .units import Units
from .utils import equals, CfList, array_allclose

class Reduction(object):

    __slots__ = ['scale_factor',
                 'add_offset',                
                 ]

    def __init__(self, **kwargs):     
        '''

Set attributes from keyword arguments.
'''
        for attr, value in kwargs.iteritems():
            setattr(self, attr, value)
    #-- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        string = ['<CF %s:' % self.__class__.__name__]
        for attr in self.__slots__:
            try:
                string.append(' %s=%s' % (attr, getattr(self, attr)))
            except AttributeError:
                pass
        #--- End: for
        string.append('>')
        return ''.join(string)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        return repr(self)
    #--- End: def

    def __nonzero__(self): 
        '''
'''
        for attr in self.__slots__:
            try:
                getattr(self, attr)
                return True
            except AttributeError:
                pass
        #--- End: for

        return False 
    #--- End: def

    def unpack(self, array):
        ''' Unpack the data *in place*, deleting any attributes used
        '''
        try:
            if self.scale_factor != 1.0:
                array *= self.scale_factor
                del self.scale_factor
        except AttributeError:
            pass
        
        try:
            if self.add_offset != 0.0:
                array += self.add_offset
                del self.add_offset
        except AttributeError:
            pass
   #--- End: def

#--- End: class


class ElementProperties(object):

    __slots__ = ['shape',          # list of lists of ints
                 'order',          # list of lists of strings
                 'direction',      # list of lists of bools
                 '_Units',         # list of Units objects
                 'reduction',      # list of Reduction objects
                 ]
    
    def __init__(self, **kwargs):
        '''

Set attributes from keyword arguments.
'''
        for attr, value in kwargs.iteritems():
            setattr(self, attr, value)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        string = ['<CF %s:' % self.__class__.__name__] 
        for attr in self.__slots__:
            try:
                string.append(" %s=%s" % (attr, repr(getattr(self, attr))))
            except AttributeError:
                pass  
        #--- End: for    
        string.append('>')
        return ''.join(string)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        return repr(self)
    #--- End: def

#--- End: class


class Data(CfList):
    '''

Store CF data array. An ordered list of arrays may be stored which,
when concatenated form the desired array.

:keyword data: Initialize the data from this object, which may be a
               :class:`~cf.Data` object or a :obj:`numpy.ndarray` or
               None. Any other type of object returns an empty
               :class:`~cf.Data` instance.

**Special attributes:**

================= ===========
Attribute         Description
================= ===========
:attr:`_list`     
:attr:`shape`     
:attr:`order`     
:attr:`direction` 
:attr:`_Units`     
:attr:`adim`      
:attr:`elements`     
================= ===========

**Methods and attributes defined here:**
'''
    __slots__ = ['_list',           # list of array-like objects
                 'shape',           # tuple of ints
                 'order',           # list of strings
                 'direction',       # dict
                 '_Units',          # Units instance
                 '_FillValue',      # _FillValue
                 'adim',            # int
                 'elements',
                 ]
    
    def __init__(self, data=None, units=Units()):
        '''

Set data and metadata from the given argument.
'''
        try:
            self.shape = data.shape
        except AttributeError:
            # 'data' is not a Data object nor an array-like object
            self._list     = []
            self.shape     = None
            self.direction = None
            self.order     = None
            self._Units    = copy.deepcopy(units)
            self.elements  = []
        else:
            try:
                # 'data' is another Data object
                self._list     = data._list[:]
                self._Units    = copy.deepcopy(data._Units)
                self.order     = copy.deepcopy(data.order)
                self.direction = copy.deepcopy(data.direction)
                self.elements  = copy.deepcopy(data.elements)
                try:
                    self.adim = data.adim
                except AttributeError:
                    pass
            except AttributeError:
                # 'data' is an array-like object, such as numpy array
                self._list     = [data]
                self._Units    = copy.deepcopy(units)
                self.order     = [str(d) for d in xrange(data.ndim)]
                self.direction = dict(((d, True) for d in self.order))
                if not self.direction:
                    self.direction = True
                    edirection     = True
                else:
                    edirection = [True] * data.ndim
                self.elements  = [ElementProperties(shape     = list(data.shape),
                                                    direction = edirection,
                                                    order     = self.order[:],
                                                    _Units    = copy.deepcopy(units),
                                                    reduction = Reduction(),
                                                    )]
            #--- End: try
        #--- End: try
    #--- End: def

    def insert(self, index, other):
        '''

D.insert(index, object) -- insert Data object before index
'''
        
        if len(other) != 1:
            raise ValueError, "Can't insert %s object with more than 1 element for now. DCH." % \
                self.__class__.__name__

        for attr in ('_list', 'elements'):
            getattr(self, attr).insert(index, getattr(other, attr)[0])
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        new = type(self)()

        # Make sure that the index is a slice
        if isinstance(index, (int, long)):
            if index == -1:                    
                index = slice(index, None, 1)
            else:
                index = slice(index, index+1, 1)            
        #--- End: if

        # Set all of the attributes except shape and adim
        new._list     = self._list[index]
        new.order     = self.order
        new.direction = self.direction
        new._Units    = self._Units
        new.elements  = self.elements[index]
        
        if not new._list:
            raise IndexError, "%s index out of range" % self.__class__.__name__

        # Set the shape and adim attributes
        len_self = len(self)
        len_new  = len(new)
#        if len_new == 1:
#            new.shape = new.elements[0].shape[:]
#            new.order = new.elements[0].order[:]
#            new.direction = dict(((o, d) for o, d in zip(new.order, 
#                                                         new.elements[0].direction)))
#
        if len_new == 1:
            shape = list(self.shape)
            e = new.elements[0]
            if len(e.shape) < self.ndim:
                new.adim = self.adim
                shape[new.adim] = 1
            else:
#                dim = new.order[adim]
                shape = e.shape
            new.shape = tuple(shape)
#            else:

#        else:
 #        if len_new == 1:
#            new.shape = self.shape
#            if new.elements[0].ndim < new.ndim:
#                new.adim = self.adim            
        elif len_new == len_self:
            #        if len_new == len_self:
            new.shape = self.shape
            try:
                new.adim = self.adim
            except AttributeError:
                pass
        else:
            adim  = self.adim
            ndim  = self.ndim
            dim   = self.order[adim]
            shape = list(self.shape)
#            if len_new == 1:
#                e = new.elements[0]
#                if len(e.shape) < ndim:
#                    shape[adim] = 1
#                else:
#                    shape[adim] = e.shape[e.order.index(dim)]
#            else:
            # Assign adim
            new.adim = adim
            a = 0
            for e in new.elements:
                if len(e.shape) < ndim:
                    a += 1
                else:
                    a += e.shape[e.order.index(dim)]
            shape[adim] = a
        #--- End: if        

            # Assign shape
            new.shape = tuple(shape)
        #--- End: if

        return new
    #--- End: def
        
    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index] = value
'''
        # Make sure that the index is a slice. This allows properly
        # assign from a single element value E.g. (for lists):
        # >>> a = [1,2,3,4]
        # >>> a[slice(1,2)] = ['X']
        # >>> a
        # [1, 'X', 3, 4]        
        # >>> a[slice(2,4)] = ['Y']
        # >>> a
        # [1, 'X', 'Y']        
        if isinstance(index, (int, long)):
            if index == -1:                    
                index = slice(index, None, 1)
            else:
                index = slice(index, index+1, 1)
        #--- End: if

        self._list[index]    = value._list
        self.elements[index] = value.elements
        
        if len(self) == 1:
            try:
                del self.adim
            except AttributeError:
                pass
    #--- End: def
        
    def __deepcopy__(self, memo):
        '''
    ''' 
        return self.copy()
    #--- End: def

    @property
    def slice(self):
        '''

Return an object which, when indexed, returns a slice of the data.
'''
        return SliceData(self)
    #--- End: def

    @property
    def size(self):
        '''
Number of elements in the data

:returns: A positive integer
    '''
#        i=1
#        for j in self.shape:
#            i *= j
#        return i
        return reduce(mul, self.shape, 1) 
    #--- End: def

    @property
    def dtype(self):
        '''

Data-type of the data's elements.

:returns:     A numpy dtype object.
'''
        try:
            return self._list[0].dtype
        except IndexError:
            raise AttributeError, "___(*__"
    #--- End: def

    @property
    def ndim(self):
        '''
        
Number of dimensions of the data.

Equivalent to the number of elements in shape tuple.

:returns:     A non-negative integer
'''
        return len(self.shape)
    #--- End: def

    def equals(self, other, rtol=None, atol=None):
        '''

Return True if the data instance is congruent to another one in that

    1. They have identical data array. Pointers to data on disk may be
       converted to numpy arrays in memory if required.

Equality of numbers is to within a tolerance. Refer to :mod:`cf` for
details.

:argument other: The variable to compare against for equality.

:type other: object
             
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the absolute tolerance to this value for all such
               comparisons.

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the relative tolerance to this value for all such
               comparisons.

:type rtol: None or float

:returns: True if the two objects are congruent, False otherwise.


**Examples:**

>>> y = x
>>> x.equals(y)
True
'''
        # Check each instance's id
        if id(self) == id(other):
            return True

        # Check that each instance is the same type
        if type(self) != type(other):
            return False
                    
        if self.shape != other.shape:
            return False

        # Compare data objects which are the same
        if len(self) == len(other):
            same_objects = True
            for data0, data1 in zip(self._list, other._list):
                if id(data0) != id(data1):
                    same_objects = False
                    break
            #--- End: for
            if same_objects:
                return True
        #--- End: if

        if not array_allclose(self.varray, other.varray,
                              rtol=rtol, atol=atol):
            return False

        return True            
    #--- End: def

#    def transpose(self, 

    def remove_dimension(self, index):
        '''

Remove a size 1 dimension fromt the data **IN PLACE**.
'''
        # Remove the dimension's size
        shape = list(self.shape)
        try:
            if shape.pop(index) != 1:
                raise ValueError, "Can't remove a dimension of size > 1"
            self.shape = tuple(shape)
        except IndexError:
            raise ValueError, "Can't remove a dimension from a scalar"
        
        # Remove the dimension for this index
        dim = self.order.pop(index)
        
        # Remove the dimension's direction
        self.direction.pop(dim)
        

        for e in self.elements:
            try:
                index = e.order.index(dim)
            except ValueError:
                # dim is not in the e.order list
                continue

            # Still here? Then dim is in the e.order list
            e.order.pop(index)
            e.shape.pop(index)
            e.direction.pop(index)                
        #--- End: for

        return dim
    #--- End: def
                        
    def conform(self, elements=None):
        '''

Conform data according to its metadata.        
'''
        if elements is None:
            # By default, conform all elements
            elements = xrange(len(self))
        else:
            try:
                elements.__iter__
            except AttributeError:
                # Conform a single element
                elements = (elements,)
        #--- End: if

        for n in elements:
            array   = self._list[n]
            element = self.elements[n]

            # --------------------------------------------------------
            # Read data from disk into a numpy array, if required
            # --------------------------------------------------------
            if not isinstance(array, numpy.ndarray):
                array = array[...]

            # --------------------------------------------------------
            # If the number of array dimensions differs from those
            # specified in the element's metadata then reshape the
            # array.
            #
            # For example, this could occur if array.shape=(1,2,1,1)
            # but previously we had decided that we didn't want the
            # first size 1 dimension. At that time, we changed
            # self.shapes[n] from [1,2,1,1] to [2,1,1], reflecting the
            # fact that this is how we want to view the data from then
            # on.
            # --------------------------------------------------------
            shape = element.shape
            if len(array.shape) != len(shape):
                array = array.reshape(tuple(shape))

            # --------------------------------------------------------
            # Reverse axes' directions
            # --------------------------------------------------------
            order     = element.order
            direction = element.direction
            full_slice = [slice(None)] * array.ndim
            for i, dim in enumerate(order):
                if direction[i] != self.direction[dim]:
                    full_slice[i] = slice(None, None, -1)
            array = array[tuple(full_slice)]
            
            # --------------------------------------------------------
            # Reorder axes
            # --------------------------------------------------------
            new_order = []
            for dim in self.order:
                try:
                    new_order.append(order.index(dim))
                except ValueError:
                    pass            
            array = numpy.transpose(array, axes=new_order)
            
            # --------------------------------------------------------
            # Unpack the data
            # --------------------------------------------------------
            try:
                element.reduction.unpack(array)
            except AttributeError:
                pass

            # --------------------------------------------------------
            # Change units
            # --------------------------------------------------------
            if element._Units != self._Units:
                raise NotImplementedError, "Haven't sorted out Units conversions, yet. DCH."

            # --------------------------------------------------------
            # Insert a missing aggregating axis
            # --------------------------------------------------------
            if array.ndim < self.ndim:
                array = numpy.expand_dims(array, axis=self.adim)        
                
            # --------------------------------------------------------
            # The array now conforms with self.shape and
            # self.order, so update the element in self
            # --------------------------------------------------------
            self._list[n]     = array
            element.shape     = list(array.shape)
            element.order     = self.order[:]
            element.direction = [self.direction[dim] for dim in self.order]
            if not element.direction:
                element.direction = self.direction
        #--- End: for

    #--- End: def

    def copy(self, data=True):
        '''

Return a deep copy.

:keyword data: *Optional*. If False then assign the output variable's
               data arrays to None, with the expectation that the data
               will be replaced subsequently. The data object's
               metadata are deep copied in any case.

:type data: bool

:returns: A deep copy.
'''
        new = type(self)()

        # Copy the data
        if data:
            for d in self._list:
                if isinstance(d, netCDF4.Variable):
                    new._list.append(d)
                else:
                    new._list.append(copy.deepcopy(d))
        else:
            new._list = [None] * len(self)
        #--- End: if

        # Deep copy the attributes
        for attr in set(self.__slots__) - set(('_list',)):
            try:
                setattr(new, attr, copy.deepcopy(getattr(self, attr)))
            except AttributeError:
                # This attribute hasn't been set
                pass
        #--- End: for

        # Return the new Data instance
        return new
    #--- End: def

    def concatenate(self, *elements):
        '''

Concatenate elements into a Data object with a single element.
'''
        self.conform(*elements)

        if len(self) == 1:
            return self

        if not elements:
            elements = slice(None)
            new      = self[0]
        else:
            try:
                if len(elements[0]) == 1:
                    new = self[elements[0][0]]
                    del new.adim
                    return new
            except TypeError:
                elements = slice(elements[0], elements[0]+1)
                new      = self[elements[0]]
            else:
                elements = slice(elements[0][0], elements[0][-1]+1)
                new      = self[elements[0][0]]                
        #--- End: if
                
            
        array = numpy.ma.concatenate(self._list[elements], axis=self.adim)

        # If none of the input numpy arrays were masked then convert
        # to a non-masked numpy array
        if array.mask is numpy.bool_(False):
            array = numpy.asarray(array)
        else:
            try:
                array.set_fill_value(self._FillValue)
            except ArttributeError:
                pass

        new                   = self[0]
        new._list             = [array]
        new.shape             = array.shape
        new.elements[0].shape = list(new.shape)

        try:
            del new.adim
        except AttributeError:
            pass

        return new
    #--- End: def

    @property
    def varray(self):
        '''

Create a numpy array view of the concatenated data.
'''
        self[:] = self.concatenate()
        return self._list[0].view()
    #--- End: def
  

    @property
    def array(self):
        '''

Create a numpy array copy of the concatenated data.
'''
        self[:] = self.concatenate()
        return self._list[0].copy()
    #--- End: def
      
#   def first_and_last_data(self, n):
#       ''' 
#
#eturn the first and last elements of the data without replacing a file pointer
#ith a numpy array.
#
#returns: The scalar value of the last element of the data.
#
#   '''
#
#       n = 0
#
#       other_end = -1
#
#       array_shape = list(self._list[0].shape)
#       shape       = self.shapes[0]
#
#       if array_shape == () or shape == ():
#           index = [0] * len(array_shape)        
#           out.append(self._list[0][tuple(index)])
#           index = numpy.where(index, 0, -1)
#           out.append(self._list[-1][tuple(index)])
#           return index
#
#       index = [0] * len(shape)
#
#       for i, dim in enumerate(self.order[0]):
#           if self.direction[0][i] != self.direction[dim]:
#               index[i] = -1
#       
#       if len(array_shape) < len(shape):
#           j=0
#           while len(array_shape) < len(shape):
#               for i in xrange(j, len(array_shape)):
#                   if array_shape[i] != shape[i]:
#                       shape.pop(i)
#                       index.pop(i)
#                       j = i + 1
#                       break
#           #--- End: while
#       elif len(shape) < len(array_shape):
#           j=0
#           while len(shape) < len(array_shape):
#               for i in xrange(j, len(shape)):
#                   if shape[i] != array_shape[i]:
#                       shape.insert(i, 1)
#                       index.insert(i, n)
#                       j = i + 1
#                       break
#           #--- End: while
#       #--- End: if
#
#       other_end = 0
#
#       out = self._data._list[n][tuple(index)]
#
#       self._data.conform((0, -1))
#
#       out = [self.first_datum(raw=True), self.last_datum(raw=True)]
#
#       if not raw and self.istime():
#           try:
#               calendar = self.calendar
#           except:
#               calendar = 'standard'
#           out = cf.Time(out, units=self.units, calendar=calendar)        
#       #--- End: if
#
#       return out
#   #--- End: def


    def collapse(self, dimensions, method='mean', weights=None):
        '''
'''      

        def sum_of_weights(x, w):
            if weights is None:
                return x.count()
            return numpy.ma.sum(numpy.ma.masked_array(w, mask=x.mask))
        #--- End: def

        def sum_of_squares_of_weights(x, w):
            if weights is None:
                return x.count()
            return numpy.ma.sum(numpy.ma.masked_array(w, mask=x.mask)**2)
        #--- End: def

        def mean(x, w, returned=False):
            # Arithmetic mean
            return numpy.ma.average(x, weights=w, returned=returned)
        #--- End: def
        
        def geometric_mean(x, w):
            # Geometric mean
            sum_w = sum_of_weights(x, w)
            return numpy.ma.exp(numpy.ma.sum(numpy.ma.log(x) * w) / sum_w)
        #--- End: def

        def harmonic_mean(x, w):
            # Harmonic mean
            sum_w = sum_of_weights(x, w)
            return sum_w / numpy.ma.sum(w/x)
        #--- End: def

        def biased_var(x, w):
            # Biased variance
            if x.count() < 2:
                raise ValueError, "dd1 dch"
            xbar, sum_w = mean(x, w, returned=1)
            return numpy.ma.sum(w*(x-xbar)**2) / sum_w
        #--- End: def
        
        def unbiased_var(x, w):
            # Unbiased variance              
            if x.count() < 2:
                raise ValueError, "dd2 dch"
            xbar, sum_w = mean(x, w, returned=1)
            sum_w2 = sum_of_squares_of_weights(x, w)
            return numpy.ma.sum(w*(x-xbar)**2) * sum_w / (sum_w**2 - sum_w2)
        #--- End: def
        
        def biased_sd(x, w):
            # Biased standard deviation
            return biased_var(x, w)**0.5
        #--- End: def
        
        def unbiased_sd(x, w):
            # Unbiased standard deviation
            return unbiased_var(x, w)**0.5
        #--- End: def
        
        def maximum(x, w):
            # 
            return numpy.ma.max(x)
        #--- End: def
        
        def minimum(x, w):
            # 
            return numpy.ma.min(x)
        #--- End: def
        
        def mid_range(x, w):
            # 
            return numpy.ma.ptp / 2.0
        #--- End: def
        
        methods = {'mean'        : mean,
                   'biased_var'  : biased_var,
                   'unbiased_var': unbiased_var,
                   }
        
        new = self.concatenate()
        a = new.varray

        new_shape = list(a.shape)
        for i in dimensions:
            new_shape[i] = 1

        # Move collapse dimensions to the
        # end. E.g. array.shape=(2,3,4,5), dimensions=(3,1) ->
        # array.shape=(2,4,5,3). And reshape -> array.shape=(8,5,3)
        len_dim = len(dimensions)
        ndim = array.ndim
        while dimensions:
            dim = dimensions.pop(0)
            a = numpy.rollaxis(a, axis=dim, start=ndim)
            for i, d in enumerate(dimensions):
                if d > dim:
                    dimensions[i] -= 1
        #--- End: while

        new_size = reduce(mul, a.shape[:ndim-len_dim], 1)
        a = a.reshape((new_size,) + a.shape[ndim-len_dim:])
        b = numpy.empty((new_size, 1))
        
        fold = methods[method]
        for x, y in zip(a, b):
            y[0] = fold(x, weights)

        b.reshape(new_shape)

        new._list[0] = b

        new.shape             = b.shape
        new.elements[0].shape = list(b.shape)

        return new

    #--- End: def

#--- End: class


class SliceData(object):

    __slots__ = ['data']

    def __init__(self, data):
        '''

Set the data attribute
'''
        self.data = data
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        def _slice(array, index, one_at_a_time):
            '''

Slice a numpy array ('array') by 'index' either in one go
(one_at_a_time is False) or one dimension at a time (one_at_a_time is
True). The aim is for the dimensionality of new_array to always be the
same as array. Index should be a tuple with one entry per array
dimension. If an entry of index is an integer then one_at_a_time
should be True.
'''
            if one_at_a_time:
                # Slice the data one dimension at a time
                full_slice   = [slice(None)] * len(index)
                new_index    = full_slice[:]
                new_index[0] = index[0]
                new_array    = array[new_index]
                for i in xrange(1, len(index)):
                    new_index    = full_slice[:]
                    new_index[i] = index[i]
                    new_array    = new_array[new_index]
                #--- End: for
                return new_array
            else:
                # Slice the data in one go
                return array[index]
        #--- End: def

        data = self.data

        index, one_dim_at_a_time = _parse_index(data, index)

        # ------------------------------------------------------------
        # If slice is Ellipsis then return a deep copy
        # ------------------------------------------------------------
        if index == (slice(None),) * data.ndim:
            data.conform()
            return data.copy()

        # Still here?    
        if len(data) == 1:

            data.conform()
            new          = type(data)(data)
            new._list[0] = _slice(new._list[0], index, one_dim_at_a_time).copy()

        else:

            adim = data.adim

            new_shape = []
            for ind, size in zip(index, data.shape):
                try:
                    start, stop, step = ind.indices(size)
                    div, mod = divmod(stop-start, step)
                    if mod != 0:
                        div += 1
                    new_shape.append(div)
                except AttributeError:
                    new_shape.append(len(ind))                       
                #--- End: try
            #--- End: for
                    
            new       = type(data)(data[0])
            new._list = [numpy.empty(tuple(new_shape), dtype=data.dtype)]

            new_index = [slice(None)] * new.ndim

            if index[adim] == slice(None):

                # ----------------------------------------------------
                # The slice over the aggregating dimension is ':', so
                # we need to slice every element of self.data
                # ----------------------------------------------------
                data.conform()

                start = 0
#                for e, array in zip(data.elements, data._list):
                for array in data._list:
                    stop            = start + array.shape[adim] #e.shape[e.order.index(data.order[adim])]
                    new_index[adim] = slice(start, stop)
                    new._list[0][tuple(new_index)] = _slice(array,
                                                            index, 
                                                            one_dim_at_a_time)
                    start = stop   
                #--- End: for
               
            else:
                
                # ----------------------------------------------------
                # If the slice over the aggregating dimension is not
                # slice(None) then we probably only need to slice some
                # elements of the list. (If slice over aggregating
                # dimension is slice(0, len(data)) then we will end up
                # slicing all elements of the list.)
                # ----------------------------------------------------

                # ----------------------------------------------------
                # Find the shape of the the output data, including
                # size 1 dimensions for now. Create an empty array of
                # that shape for storing the slices from the
                # appropiate elements of the 'data'
                # ----------------------------------------------------
               
                adim_index     = index[adim]                
                len_adim_index = data.shape[adim]
                try:
                    start, stop, adim_step = adim_index.indices(len_adim_index)
                    if adim_step < 0:
                        start, stop, adim_step = (stop+((start-stop)%(-adim_step)), 
                                                  start+1,
                                                  -adim_step)
                        reorder = slice(None, None, -1)
                except AttributeError:
                    adim_index        = numpy.array(adim_index)
                    adim_index        = numpy.where(adim_index < 0, 
                                                    adim_index+len_adim_index, adim_index)
                    adim_index_sorted = numpy.sort(adim_index)
                    start             = adim_index_sorted[0]
                    stop              = adim_index_sorted[-1] + 1
                    if (adim_index_sorted != adim_index).any():
                        reorder = numpy.argsort(numpy.argsort(adim_index))
                #--- End: try

                # ----------------------------------------------------
                # Loop round the elements of 'data', slicing those
                # elements whose aggregating dimension overlaps the
                # master slice (the input slice for the aggregating
                # dimension)
                # ----------------------------------------------------
                nth_stop  = 0
                new_start = 0

                for n, e in enumerate(data.elements):

                    # Length of the n-th element's aggregating
                    # dimension
                    try:
                        len_nth_adim = e.shape[e.order.index(data.order[adim])]
                    except ValueError:
                        len_nth_adim = 1

                    # Master index for the start of the n-th element's
                    # aggregating dimension
                    nth_start = nth_stop 
                    # Master index for the end of the n-th element's
                    # aggregating dimension (plus 1)
                    nth_stop += len_nth_adim

                    if nth_stop < start:
                        # Continue if the n-th element's aggregating
                        # dimension is entirely before the master
                        # slice
                        continue
                    if nth_start >= stop:
                        # Stop if the n-th element's aggregating
                        # dimension is entirely beyond the master
                        # slice, and therefore all further elements
                        # ((n+1)-th,...) also satisfy this
                        break

                    # ------------------------------------------------
                    # Still here? The find 'nth_index', the index for
                    # slicing the n-th element of the input data
                    # object and find 'new_index', the index for
                    # putting that slice into the only element of the
                    # output data object
                    # ------------------------------------------------
                    nth_index = list(index)
                    try:  
                        # The master slice is a slice object, so
                        # nth_index[adim] will be a slice object
                        # ('adim_step' is defined)
                        slice_step  = adim_step
                        slice_start = start - nth_start
                        slice_stop  = nth_stop

                        if slice_start < 0:
                            slice_start = slice_start % adim_step
                        if stop < nth_stop:
                            slice_stop = len_nth_adim - (nth_stop - stop)

                        nth_index[adim] = slice(slice_start, slice_stop, slice_step)
                        nth_index       = tuple(nth_index)

                        # Find the length of this index
                        slice_start, slice_stop, slice_step = \
                            nth_index[adim].indices(len_nth_adim)
                        div, mod = divmod(slice_stop-slice_start, slice_step)
                        if mod != 0:
                            div += 1
                        len_nth_index_adim = div
                    except NameError:
                        # The master slice is a sequence of integers,
                        # so nth_index[adim] will be a sequence of
                        # integers ('adim_step' is not defined)
                        temp = []
                        for i in adim_index_sorted:
                            if i >= nth_stop:
                                break
                            temp.append(i - nth_start)

                        nth_index[adim]              = temp
                        nth_index, one_dim_at_a_time = _parse_index(new, tuple(nth_index))   

                        # Find the length of this index
                        len_nth_index_adim = len(temp)
                        # update adim_ind_sorted
                        adim_index_sorted = adim_index_sorted[len(temp):]
                    #--- End: try  

                    if not len_nth_index_adim:
                        continue

                    # ------------------------------------------------
                    # Slice the n-th element of data, putting the
                    # result into the correct place in the 'new'
                    # array.
                    # ------------------------------------------------
                    # Conform this data element
                    data.conform(n)

                    # Update the slice to new
                    new_stop        = new_start + len_nth_index_adim
                    new_index[adim] = slice(new_start, new_stop)
                    new_start       = new_stop

                    new._list[0][tuple(new_index)] = _slice(data._list[n], 
                                                            nth_index, 
                                                            one_dim_at_a_time)          
                #--- End: for

                # ----------------------------------------------------
                # If the master slice for the aggregating dimension
                # was a not an ascending sequence of indices, then we
                # had to reorder it so that it was. In this case, the
                # new._list[0] array is currently in the wrong order
                # along the aggregating dimension and so will need to
                # be reordered.
                # ----------------------------------------------------
                try:                    
                    # Reorder indices were defined because the master
                    # slice for the aggregating dimension is not an
                    # ascending sequence
                    new_index[adim] = reorder 
                    new_index, one_dim_at_a_time = _parse_index(new, tuple(new_index)) 

                    new._list[0] = _slice(new._list[0],
                                          new_index, 
                                          one_dim_at_a_time)
                except NameError:
                    # No reorder indices were defined because the
                    # master slice for the aggregating dimension is
                    # ascending sequence
                    pass

            #--- End: if
        #--- End: try

        # If none of the input numpy arrays were masked then convert
        # to a non-masked numpy array
        try:
            if new._list[0].mask is numpy.bool_(False):
                new._list[0] = numpy.asarray(new._list[0])
        except AttributeError:
            pass

        new.shape                 = new._list[0].shape
        new.elements[0].shape     = list(new.shape)
        new.elements[0].order     = new.order[:]
        new.elements[0].direction = [new.direction[dim] for dim in new.order]
        if not new.elements[0].direction:
            new.elements[0].direction = new.direction

#        print 'new.elements = ', new.elements
#        print 'new.order, dir = ', new.order, new.direction

  
        return new
    #--- End: def

    def __setitem__(self, index, value):
        '''

x.__setitem__(index, value) <==> x[index].value
'''
        # Don't do this like in __getitem__ (which is less memory and
        # CPU intensive) for now, since I don't want to chop up
        # 'value' to match individual elements of 'self'. For example,
        # if data.shape=(2,2) then data[:,:]=5, data[:,:]=[5] and
        # data[:,:]=[5,5] are all valid assignments.

        data = self.data

        index, one_dim_at_a_time = _parse_index(data, index)

        if len(data) == 1:
            # -----------------------------------------------------
            # self.data has only one element, so every element of
            # self.data is assigned to
            # -----------------------------------------------------
            data.varray[index] = value
            return None          


#        try:
#            adim = data.adim
#        except AttributeError:
#            # -----------------------------------------------------
#            # self.data has only one element, so every element of
#            # self.data is assigned to
#            # -----------------------------------------------------
#            data.varray[index] = value
#            return None          

        # Still here?

        raise NotImplementedError, "Assigning to aggregated fields is not yet implemented"
        adim = data.adim
        index_adim = index[adim]

        if index_adim == slice(None):
            # -----------------------------------------------------
            # The aggregating dimension's slice is ':', so every
            # element of self.data is assigned to
            # -----------------------------------------------------
            data.varray[index] = value
            return None

        # Still here?

        # ----------------------------------------------------
        # If the slice over the aggregating dimension is not
        # slice(None) then we probably only need to slice some
        # elements of the list. (If slice over aggregating
        # dimension is slice(0, len(data)) then we will end up
        # slicing all elements of the list.)
        # ----------------------------------------------------                

        # -----------------------------------------------------------
        # Find the first (start) and last (stop) indices of the slice
        # along the aggregating dimension. This will allow us to find
        # which elements of self.data need assigning to.
        # -----------------------------------------------------------
        try:
            start, stop, step = index_adim.indices(data.shape[adim])
        except AttributeError:
            index_adim = numpy.where(index_adim < 0, 
                                     index_adim+data.shape[adim], index_adim)
            start = numpy.amin(index_adim)
            stop  = numpy.amax(index_adim) + 1

        # ------------------------------------------------------------
        # Loop round the elements of self.data, finding out which ones
        # need their aggregating dimension assigned to. Remember these
        # elements via their indices of self.data.
        # ------------------------------------------------------------
        elements = []
        nth_stop  = 0
        for n, shape in enumerate(data.shapes):

            # --------------------------------------------------------
            # Conform the element's shape to find the size of its
            # aggregating dimension.
            # --------------------------------------------------------
            new_order = []
            for dim in self.order:
                try:
                    new_order.append(self.order[n].index(dim))
                except ValueError:
                    pass            
            shape = [shape[n] for n in new_order]
            if len(shape) < len(self.shape):
                shape.insert(adim, 1)

            nth_start = nth_stop 
            nth_stop += shape[adim]

            if nth_stop < start:
                continue
            if nth_start >= stop:
                break

            elements.append(n)
            
            try:
                offset
            except NameError:
                # Name 'offset' is not defined
                offset = nth_start
        #--- End: for 
                
        # --------------------------------------------------------
        # Concatenate the elements being assigned to into a single
        # numpy array
        # --------------------------------------------------------
        data[elements] = data.concatenate(elements)

        # --------------------------------------------------------
        # Assign to the concatenated numpy array
        # --------------------------------------------------------
        try:
            index[adim] = slice(start-offset, stop-offset, step)
        except NameError:
            # Name 'step' is not defined
            index[adim] = index_adim - offset
        data[elements[0]].varray[index] = value
    #--- End: def

#--- End: class


def set_Data_attributes(f, key, v):
    '''

Set the metadata.    
    '''       
    # ----------------------------------------------------------------
    # Axis order and direction
    # ----------------------------------------------------------------
    if not v._data.ndim:
        # Scalar data       
        v._data.order = []
        try:
            # The variable is a dimension coordinate
            v._data.direction = f.space.increasing[key]
        except KeyError:
            # The variable is not a dimension coordinate
            v._data.direction = True
        v._data.elements[0].direction = v._data.direction
    else:
        # 1-d or N-d dimensional data
        v._data.order     = f.space.dimensions[key][:]        
        v._data.direction = dict((dim, f.space.increasing[dim]) for dim in v._data.order)
        v._data.elements[0].direction = [v._data.direction[dim] for dim in v._data.order]
 
    # ----------------------------------------------------------------
    # Units
    # ----------------------------------------------------------------
    try:
        v._data._Units = v._private.pop('_Units')
    except (KeyError, AttributeError):
        pass

    v._data.elements[0].order  = v._data.order[:]
    v._data.elements[0]._Units = copy.deepcopy(v._data._Units)

    # ----------------------------------------------------------------
    # Reductions
    # ----------------------------------------------------------------
    for attr in ('scale_factor', 'add_offset'):
        try:
            getattr(v._data._list[0], attr)
            try:
                delattr(v, attr)
            except AttributeError:
                pass   
        except AttributeError:
            try:
                value = getattr(v, attr)
                setattr(v._data.elements[0].reduction, attr, value)
                delattr(v, attr)
            except AttributeError:
                pass            
    #--- End: for
                          
    # ----------------------------------------------------------------
    # _FillValue
    # ----------------------------------------------------------------
    try:
        v._data._FillValue = v._private.pop('_FillValue')
    except (KeyError, AttributeError):
        pass

    v._data.elements[0].order  = v._data.order[:]
    v._data.elements[0]._Units = copy.deepcopy(v._data._Units)

    # ----------------------------------------------------------------
    # If the variable is a coordinate then set the metadata for its
    # bounds' data
    # ----------------------------------------------------------------
    if isinstance(v, Coordinate):
        try:
            b = v.bounds
        except AttributeError:
            pass
        else:            
            b._data.order  = v._data.order[:] + ['bounds'] 
            # Units
            try:
                b._data._Units = b._private.pop('_Units')
            except (KeyError, AttributeError):
                pass
 
            if v.ndim == 0:
                # Scalar coordinate (note that key must be dimN because
                # only dimension coords can be scalar)
                b._data.direction             = {'bounds': v._data.direction}
                b._data.elements[0].direction = [v._data.direction]
            elif v.ndim == 1:
                # 1-d coordinate
                b._data.direction           = v._data.direction.copy()
                b._data.direction['bounds'] = v._data.direction[v._data.order[0]]
                b._data.elements[0].direction = v._data.elements[0].direction[:]
                b._data.elements[0].direction.append(b._data.direction['bounds'])
            else:
                # N-d coordinate
                b._data.direction           = v._data.direction.copy()
                b._data.direction['bounds'] = True
                b._data.elements[0].direction = v._data.elements[0].direction[:]
                b._data.elements[0].direction.append(True)
            #--- End: if
             
            b._data.elements[0].order  = b._data.order[:]
            b._data.elements[0]._Units = copy.deepcopy(b._data._Units)

            # Reductions
            for attr in ('scale_factor', 'add_offset'):
                try:
                    getattr(b._data._list[0], attr)
                    try:
                        delattr(b, attr)
                    except AttributeError:
                        pass   
                except AttributeError:
                    try:
                        value = getattr(b, attr)
                        setattr(b._data.elements[0].reduction, attr, value)
                        delattr(b, attr)
                    except AttributeError:
                        pass            
            #--- End: for
       #--- End: try
    #--- End: if

    # ----------------------------------------------------------------
    # Return the updated variable
    # ----------------------------------------------------------------
    return v
#--- End: def

def dimension_direction(coord):
    '''

Ascertain if a 1-d monotonic coordinate is ascending (True) or
desending (False).
    '''
    c = coord._data._list[0]

    if c.size > 1:
        return c[0] < c[-1]
 
    # Still here? Then infer the direction from the dimension
    # coordinate's bounds, if it has any
    try:
        b = coord.bounds._data._list[0]
    except AttributeError:
        pass
    else:  
        b = b[(0,)*coord.ndim + (slice(None),)]
        return b[0] < b[1]

    # Still here? Then infer the direction from the dimension
    # coordinate's positive attribute, if it has one
    try:
        if coord.getpub('positive').lower().startswith('d'): # d for down
            return False
    except AttributeError:
        pass

    # Still here? Then assume that the dimension is increasing
    return True   
#--- End: def
