import glob
import netCDF4
import os
import struct
from .utils     import _files
from .field     import FieldList
from .aggregate import aggregate as cf_aggregate
import netcdf
import pp

def _file_format(openfile):
    '''

Private function to read a file's magic number.

Takes a file object as input and assumes that we're currently
pointing to the beginning of the file. The file is rewound to the
beginning after reading the magic number.

Returns a string giving the format of the input file.
''' 

    if openfile.name.startswith('http://'):
        # Assume that a URL is netCDF
        return 'nc'

    # Read the magic number
    magic_number = struct.unpack('=L', openfile.read(4))[0]

    # Reset the pointer to the beginning of the file
    openfile.seek(0, os.SEEK_SET)

    # ----------------------------------------------------------------
    # netCDF
    # ----------------------------------------------------------------
    if magic_number in (21382211, 1128547841, 38159427): #1128547841
        return 'nc'

    # ----------------------------------------------------------------
    # PP
    # ----------------------------------------------------------------
    if magic_number in (256, 65536):
        return 'pp'

    # ----------------------------------------------------------------
    # Add more file formats here ...
    # ----------------------------------------------------------------

    # Still here?
    raise NotImplementedError, "Unsupported file format from %s. Magic number=%d" % \
        (openfile.name, magic_number)
#--- End: def

def _read_a_file(filename, ignore_ioerror=False, verbose=False,
                 close=False, exact=False, aggregate=True,
                 extract={}):

    # Find this input file's format
    if filename.startswith('http://'):
        format = 'nc'
    else:        
        try:
            openfile = open(filename, 'rb')             
            _files(openfile)
        except IOError as error:
            if ignore_ioerror:
                print 'Ignoring IOError: %(error)s' % \
                    locals()
                return FieldList()
            raise IOError, error
                
        format = _file_format(openfile)        
    
    # ----------------------------------------------------------------
    # Read the file into fields
    # ----------------------------------------------------------------
    if format == 'nc':
        fields = netcdf.read(filename, close=close, exact=exact, 
                             extract=extract)
        
    elif format == 'pp': 
        fields = pp.read(openfile, close=close, exact=exact,
                         extract=extract)

    # Add more file formats here ...

    # ----------------------------------------------------------------
    # Aggregate fields
    # ----------------------------------------------------------------
    if aggregate:
        cf_aggregate(fields, verbose=verbose)

    return fields
#--- End: def

def read(files, read_data=False, verbose=False, index=None, 
         ignore_ioerror=False, close=False, aggregate=True,
         calendar=None,
         long_name=None,
         standard_name=None,
         units=None,
         extract={},         
         ):
    '''

Read fields from input files from disk or from an OPeNDAP server.

By default, all data arrays from all field components are not read
from disk but are stored in the returned fields as file pointers. This
behaviour may be changed with the `read_data` parameter.

NetCDF dimension names are stored in the `nc_dimensions` attribute of
a field's space and netCDF variable names are stored in the private
attribute `ncvar` of the field and each of its space components.

Fields referenced by formula_terms transforms are not included in the
returned list of fields. A field may be referenced in the transforms
of multiple output fields, so in-place changes to that field in one
transform will be seen in the other fields which reference it.

:argument files: A string or sequence of strings giving the file names
                 or OPenDAP URLs from which to read fields. For files
                 on disk, the file names may contain wild cards as
                 understood by the :mod:`glob` module.

:type files: `str` or sequence

:parameter read_data: *Optional*. If True then read the data of each
                      field, including the data of its space
                      components into memory Otherwise all data are
                      stored as file pointers.

:type data: `bool`

:parameter index: *Optional*. If None then return all possible fields
                  from the input files. If a non-negative integer then
                  only return the field with this index in the
                  otherwise full list of fields.

:type index: `None` or `int`

:parameter verbose: *Optional*. If True then print information to
                    stdout.

:type verbose: `bool`

:parameter close: *Optional*. If True then close each file after it
                  has been read.

:type close: `bool`

:parameter exact: *Optional*. If True then field selection cirteria
                  names are considered exact, as opposed to
                  abbreviations.

:type exact: `bool`

:parameter ignore_ioerror: *Optional*. If True then ignore files which
                           produce IO errors (such as empty files, for
                           example).

:type ignore_ioerror: `bool`

:parameter aggregate: *Optional*. If True then aggregate the fields
                      read in from all input files into as few fields
                      as possible using the CF aggregation
                      rules. (Note that this code has not yet been
                      optimsed for speed or memory usage.)

:type aggregate: `bool`

:parameter calendar: *Optional*. Equivalent to setting 'calendar' in
                     the `extract` keyword dictionary. Forces the
                     `exact` keyword to be True.

:type calendar: `str`


:parameter long_name: *Optional*. Equivalent to setting 'long_name' in
                      the `extract` keyword dictionary. Forces the
                      `exact` keyword to be True.

:type long_name: `str`


:parameter standard_name: *Optional*. Equivalent to setting
                          'standard_name' in the `extract` keyword
                          dictionary. Forces the `exact` keyword to be
                          True.

:type standard_name: `str`


:parameter units: *Optional*. Equivalent to setting 'units' in the
                  `extract` keyword dictionary. Forces the `exact`
                  keyword to be True.

:type units: `str`

:parameter extract: *Optional*. A dictionary of field selection
                    criteria. Only fields which match the criteria are
                    returned. Refer to :meth:`~cf.Field.match` (if
                    `exact` is True) or :meth:`~cf.Field.ematch` (if
                    `exact` is False) for details.

:type extract: `dict`

:returns: A list of fields in a :class:`~cf.FieldList` object.

**Examples:**

>>> f = cf.read('file*.nc')
>>> type(f)
<class 'cf.field.FieldList'>
>>> f
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>,
 <CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.read('file*.nc')[0:2]
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>]

>>> cf.read('file*.nc', index=0)
[<CF Field: pmsl(30, 24)>]

>>> cf.read('file*.nc')[-1]
<CF Field: temperature_wind(17, 29, 24)>

>>> cf.read('file*.nc', extract={'units': 'K'})
[<CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.read('file*.nc', standard_name='.*pmsl*', extract={'units': 'K|Pa'})[0]
<CF Field: pmsl(30, 24)>
'''

    # Initialize the output list of fields
    field_list = FieldList()

    # Initialize the extract dictionary
    extract = extract.copy()
    if calendar is not None:
        extract['calendar']      = calendar    
    if long_name is not None:
        extract['long_name']     = long_name
    if standard_name is not None:
        extract['standard_name'] = standard_name
    if units is not None:
        extract['units']         = units

    # Count the number of fields (in all files) and the number of
    # files
    field_counter = -1
    file_counter  = 0

    if isinstance(files, basestring):
        files = (files,)

    for file_glob in files:

        if file_glob.startswith('http://'):
            # Do not glob a URL
            files2 = (file_glob,)
            url = True
        else:
            # Do glob files on disk
            files2 = glob.iglob(file_glob)
            url = False

        for filename in files2:

            # Print some informative messages
            if verbose and index is None:
                print 'File: %(filename)s' % locals()

            # --------------------------------------------------------
            # Read the file into fields
            # --------------------------------------------------------
            fields = _read_a_file(filename,
                                  ignore_ioerror=ignore_ioerror,
                                  verbose=verbose,
                                  close=close,
                                  extract=extract,
                                  aggregate=aggregate)
            
            # --------------------------------------------------------
            # Add this file's fields to those alread read from other
            # files
            # --------------------------------------------------------
            field_list.extend(fields)
   
            field_counter = len(field_list)
            file_counter += 1

            # Print some informative messages
            if verbose and index is None:
                i = field_counter - len(fields)
                for s in fields:
                    print '%d: %s' % (i, repr(s))
                    i += 1
            #--- Endf: if

            # --------------------------------------------------------
            # If we only want one field from all input files then
            # break now if we have got it
            # --------------------------------------------------------
            if index is not None and field_counter >= index+1:
                break
        #--- End: for
            
        # ------------------------------------------------------------
        # If we only one field from all input files then break now if
        # we have got it
        # ------------------------------------------------------------
        if index is not None and field_counter >= index:
            break
    #--- End: for     

    # Error check
    if not ignore_ioerror:
        if not file_counter:
            raise RuntimeError('No files found')
        if not field_list:
            raise RuntimeError('No fields found from '+str(file_counter)+' files')
    #--- End: if
    if index >= len(field_list):
        raise IndexError('FieldList index='+str(index)+' is out of range')

    # Print some informative messages
    if verbose:
        if index is None:  
            print 'Read %d field%s from %d file%s' % \
                (field_counter, ('s' if field_counter!=1 else ''),
                 file_counter , ('s' if file_counter !=1 else ''))
        else:
            print 'File: %(filename)s' % locals()
            print '%d: %s' % (0, repr(field_list[index]))
            print 'Read 1 field from 1 file'
    #--- End: if
    
    # ----------------------------------------------------------------
    # 
    # ----------------------------------------------------------------
    try:
        field_list = field_list[index]
    except TypeError:
        pass

    # ----------------------------------------------------------------
    # Aggregate the output fields
    # ----------------------------------------------------------------    
    if file_counter > 1 and aggregate:
        cf_aggregate(field_list, verbose=verbose)

    # ----------------------------------------------------------------
    # Read the data arrays
    # ----------------------------------------------------------------
    if read_data:
        field_list.slice[...]

    return field_list
#--- End: def
