import collections
import copy
import numpy 
import re
import sys
#from . import _files

# Default tolerances for numeric equality (same as numpy defaults)
_NUMERIC_TOLERANCE = {'RTOL': sys.float_info.epsilon,
                      'ATOL': sys.float_info.epsilon,
                      }

_open_files = []

# ====================================================================
#
# Utility objects
#
# ====================================================================

class CfList(collections.MutableSequence):
    '''

A list-like object (a 'cf list') suitable for CF structures.
 
:argument sequence: *Optional*. Initialize new list from
                    sequence's items.

:type sequence: iterable

**Overloaded operators:**

The ``in`` (set membership), ``==`` and ``!=`` operators are
overloaded to use numerically tolerant equality.

The ``+, *, +=`` and ``*=`` operators behave as for a built-in
list.

**Methods:**

========================== =======================================================
Method                     Description
========================== =======================================================
:meth:`~cf.CfList.append`  As for a built-in list.
:meth:`~cf.CfList.copy`    Create a deep copy.
:meth:`~cf.CfList.count`   As for a built-in list but using numerically tolerant
                           equality.
:meth:`~cf.CfList.equals`  Determine whether two lists are congruent element-wise.
:meth:`~cf.CfList.extend`  As for a built-in list.
:meth:`~cf.CfList.index`   As for a built-in list but using numerically tolerant
                           equality.
:meth:`~cf.CfList.insert`  As for a built-in list.
:meth:`~cf.CfList.pop`     As for a built-in list.
:meth:`~cf.CfList.remove`  As for a built-in list.
:meth:`~cf.CfList.reverse` As for a built-in list.
:meth:`~cf.CfList.sort`    As for a built-in list.
========================== =======================================================

**Methods and attributes defined here:**
'''

    def __init__(self, sequence=()):
        '''

Set the built-in from the argument.
'''
        self.__dict__['_list'] = []
        try:
            self.__dict__['_list'].extend(sequence)
        except TypeError:
            self.__dict__['_list'].append(sequence)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return repr(self._list)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        return str(self._list)
    #--- End: def

    def __contains__(self, item):
        '''
x.__contains__(y) <==> y in x

Uses numerically tolerant equality.
'''
        for s in self._list:
            if equals(item, s):
                return True
        #--- End: for               
        return False
    #--- End: def
    
    def __len__(self):
        '''
x.__len__() <==> len(x)
'''
        return len(self._list)
    #--- End: def

    def __delitem__(self, index):
        '''
x.__delitem__(index) <==> del x[index]
'''
        del self._list[index]
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        out = self._list[index]
        try:
            int(index)
        except TypeError:
            # index is a slice instance, so return a CfList instance
            out = type(self)(out)

        return out
    #--- End: def

    def __setitem__(self, index, value):        
        '''
x.__setitem__(index, value) <==> x[index]=value
'''
        try:
            self._list[index] = value._list
        except AttributeError:
            self._list[index] = value
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y

Uses numerically tolerant equality.
'''
        return self.equals(other)
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y

Uses numerically tolerant equality.
'''
        return not self.__eq__(other)
    #--- End: def

    def __add__(self, other):
        '''
x.__add__(y) <==> x+y
        '''
        new = self[:]
        new.extend(other)
        return new
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(n) <==> x*n
        '''
        return type(self)(self._list*n)
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(n) <==> n*x
        '''
        return self.__mul__(other)
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
        '''
        self.extend(other)
        return self
    #--- End: def

    def __imul__(self, other):
        '''
x.__imul__(n) <==> x*=n
        '''
        self._list = self._list*n
        return self
    #--- End: def

    def count(self, value):
        '''
CFL.count(value) -> integer -- return number of occurrences of value

Uses numerically tolerant equality.
    '''
        rtol = RTOL()
        atol = ATOL()

        return sum(1 for v in self._list 
                   if equals(value, v, rtol=rtol, atol=atol))
    #--- End def

    def index(self, value, start=0, stop=None):
        '''

CFL.index(value, [start, [stop]]) -> integer -- return first index of value

Uses numerically tolerant equality.
'''
        rtol = RTOL()
        atol = ATOL()
       
        if start < 0:
            start = len(self) + start
        if stop is None:
            stop = len(self)
        elif stop < 0:
            stop = len(self) + stop

        for i, x in enumerate(self[start:stop]):
            if equals(value, x, rtol=rtol, atol=atol):
               return i + start
        #--- End: for

        # Still here?
        raise ValueError, "%s.index(x): x not in %s" % \
            (self.__class__.__name__, self.__class__.__name__)
    #--- End: def

    def insert(self, index, object):
        '''

CFL.insert(index, object) -- insert object before index
    '''
        try:     
            self._list.insert(index, object._list)
        except AttributeError:
            self._list.insert(index, object)
    #--- End: def

    def copy(self):
        '''

CFL.copy() -> a deep copy of CFL
'''
        return copy.deepcopy(self)
    #--- End: def

    def equals(self, other, rtol=None, atol=None):
        '''

Return True if two lists are congruent in that

    1. Each pair of their elements are equal.
    2. Both instances have equal attributes.   

Numerical equality is to within a tolerance. Refer to :mod:`cf` for
details.

:argument other: The variable to compare against for equality.

:type other: object
                                                 
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the absolute tolerance to this value for all such
               comparisons.

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the relative tolerance to this value for all such
               comparisons.

:type rtol: None or float


:returns: True if the two objects are congruent, False otherwise.

    '''
        if id(self) == id(other):
            return True
         
        # Check that each instance is the same type
        if type(self) != type(other):
            return False

        # Check that the lists have the same number of elements
        if len(self) != len(other):
            return False
               
        # Check the attributes
        attrs = set(self.__dict__)
        if attrs != set(other.__dict__):
            return False

        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()

        for attr, value in attrs - set(('_list',)):
            x = getattr(self, attr)
            y = getattr(other, attr)
            if not equals(x, y, rtol=rtol, atol=atol):
                return False
        #--- End: for

        # Check the element values
        for x, y in zip(self._list, other._list):
            if not equals(x, y, rtol=rtol, atol=atol):
                return False
        #--- End: for

        return True
    #--- End: def
    
#--- End: class


class CfDict(collections.MutableMapping):
    '''

A dictionary-like object (a 'cf dictionary') suitable for storing CF
structures.

:argument *args: Keys and values are initialized exactly as for a
                 built-in dict.

:argument **kwargs: Keys and values are initialized exactly as for a
                    built-in dict.

**Overloaded operators:**

The ``in`` (set membership) operator is overloaded to use numerically
tolerant equality.

**Methods:**

============================= ==================================================
Method                        Description
============================= ==================================================
:meth:`~cf.CfDict.clear`      As for a built-in dict.
:meth:`~cf.CfDict.copy`       Create a deep copy.
:meth:`~cf.CfDict.equals`     Determine whether two instances are congruent with
                              each other.
:meth:`~cf.CfDict.get`        As for a built-in dict.
:meth:`~cf.CfDict.get_keys`   Return keys which match a pattern.
:meth:`~cf.CfDict.has_key`    As for a built-in dict but using numerically
                              tolerant equality.
:meth:`~cf.CfDict.items`      As for a built-in dict.
:meth:`~cf.CfDict.iteritems`  As for a built-in dict.
:meth:`~cf.CfDict.iterkeys`   As for a built-in dict.
:meth:`~cf.CfDict.itervalues` As for a built-in dict.
:meth:`~cf.CfDict.keys`       As for a built-in dict.
:meth:`~cf.CfDict.pop`        As for a built-in dict.
:meth:`~cf.CfDict.popitem`    As for a built-in dict.
:meth:`~cf.CfDict.setdefault` As for a built-in dict.
:meth:`~cf.CfDict.update`     As for a built-in dict.
:meth:`~cf.CfDict.values`     As for a built-in dict.
============================= ==================================================


**Methods and attributes defined here:**
'''
    def __init__(self, *args, **kwargs):
        '''

Initialize the built-in dictionary from the arguments.
'''

        self._dict = dict(*args, **kwargs)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return repr(self._dict)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        return str(self._dict)
    #--- End: def

    def __getitem__(self, key):
        '''

Implement indexing of the cf dictionary.

:returns: A cf dictionary key's value.
'''     
        return self._dict[key]
    #--- End: def

    def __setitem__(self, key, value):
        '''
x.__setitem__(key, value) <==> x[key]=value
'''
        self._dict[key] = value
    #--- End: def

    def __delitem__(self, key):
        '''
x.__delitem__(key) <==> del x[key]
'''
        del self._dict[key]
    #--- End: def

    def __iter__(self):
        '''
x.__iter__() <==> iter(x)
'''
        return iter(self._dict)
    #--- End: def

    def __len__(self):
        '''
x.__len__() <==> len(x)
'''
        return len(self._dict)
    #--- End: def

#    def __contains__(self, item):
#        '''
#Test for set membership using numerically tolerant equality.
#'''
#        for s in self:
#            if equals(item, s):
#                return True
#
#        return False
#    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
'''
        return self.equals(other)
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y
'''
        return not self.__eq__(other)
    #--- End: def

    def has_key(self, key):
        '''
CFD.has_key(k) -> True if CFD has a key k, else False

'''
        return self._dict.has_key(key)
    #--- End: def

    def copy(self):
        '''

CFD.copy() -> a deep copy of CFD
'''
        return copy.deepcopy(self)
    #--- End: def

    def equals(self, other, rtol=None, atol=None):
        '''

Return True if two instances are congruent in that they have

    1. Equal sets of keys.    
    2. For each key, the two values are equal.
    3. Both instances have equal attributes.
    
Equality of numbers is to within a tolerance. Refer to :mod:`cf` for
details.

:argument other: The variable to compare against for equality.

:type other: object
                                                 
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for equality of real
               numbers (refer to :mod:`cf` for details). If a float
               then set the absolute tolerance to this value for all
               comparisons (refer to the :obj:`override` parameter).

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for equality of real
               numbers (refer to :mod:`cf` for details). If a float
               then set the relative tolerance to this value for all
               comparisons (refer to the :obj:`override` parameter).

:type rtol: None or float


:returns: True if the two objects are congruent, False otherwise.

'''
        if id(self) == id(other):
            return True
        
        # Check that each instance is the same type
        if type(self) != type(other):
            return False
               
        # Check the attributes
        attrs = set(self.__dict__)
        if attrs != set(other.__dict__):
            return False

        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()

        for attr in attrs - set(('_dict',)):
            x = getattr(self, attr)
            y = getattr(other, attr)
            if not equals(x, y, rtol=rtol, atol=atol):
                return False
        #--- End: for

        # Check that the keys are equal
        if set(self) != set(other):
            return False

        # Check that the key values are equal
        for key in self:
            if not equals(self[key], other[key], rtol=rtol, atol=atol):
                return False
                
        # Still here?
        return True
    #--- End: def

    def get_keys(self, regex=None):
        '''

Return a list of the cf dictionary's key names which match the given
regular expression.

:keyword regex: *Optional*. The regular expression with which to
                identify key names.

:type regex: str


:returns: A list of keys names.

**Examples:**

>>> d.keys()
['dim2', 'dim0', 'dim1', 'aux0', 'cm0']
>>> d.get_keys()
['dim2', 'dim0', 'dim1', 'aux0', 'cm0']
>>> d.get_keys('dim')
['dim2', 'dim0', 'dim1']
>>> d.get_keys('^aux|^dim')
['dim2', 'dim0', 'dim1', 'aux0']
>>> d.get_keys('dim[123]')
['dim2', 'dim1']
'''
        if regex is None:
            return self.keys()

        keys = []        
        for key in self:
            if re.search('%s' % regex, key):
                keys.append(key)
        
        return keys
    #--- End: def

#--- End class


class Comparison(object):
    '''

Create an object for storing a comparison expression.

A comparison expression comprises a relation (such as 'less than') and
a value (any object capable of being related to with the standard
comparison operators). These are stored in the `relation` and `value`
attributes respectively.

The comparison is evaluated for an arbitrary object with the
:meth:`~cf.Comparison.evaluate` method.

Valid relations are:

========  =============================================
Relation  Description
========  =============================================
lt        Strictly less than
le        Less than or equal
gt        Strictly greater than
ge        Greater than or equal
eq        Equals (within a tolerance)
ne        Not equal to (within a tolerance)
inside    Inside a given range (range bounds included)
outside   Outside a given range (range bounds excluded)
========  =============================================

.. seealso:: As a convenience, for each relation in the above list
             there is an identically named function which returns the
             appropriate :class:`~cf.Comparison` object.

:keyword **kwargs: A keyword argument to the call is the relation of
                   the comparison (any one of the relations listed
                   above). The keyword's argument is the object to
                   test the comparison against. If the keyword
                   argument is "inside" or "outside" then the value
                   should be a two-element list or tuple.

**Special attributes:**

================ =======================================
Attribute        Description
================ =======================================
:attr:`relation` A string definging the comparison
                 operator. One of relations given above.

:attr:`value`    The object with which to compare.
================ =======================================

**Examples:**

>>> c = cf.Comparison(le = 5)
>>> c.evaluate(4)
True
>>> c.evaluate(5)
True

>>> c = lt(5)
>>> c.evaluate(4)
True

>>> c = cf.Comparison(inside = (1,2))
>>> a = numpy.arange(4)
>>> c.evaluate(a)
array([False,  True,  True, False], dtype=bool)
>>> (a >= 1) & (a <= 2)
array([False,  True,  True, False], dtype=bool)


**Methods and attributes defined here:**
'''

    def __init__(self, **kwargs):
        '''

Set the attributes from the arguments.
'''
        for key, value in kwargs.items():
            self.relation = key
            self.value    = value
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return '<CF '+str(self)+'>'
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
''' 
        return 'Comparison: x '+self.relation+' '+repr(self.value)
    #--- End: def

    def dump(self, id=None, omit=()):
        '''
        
Return a string containing a full description of the object.

:keyword id: *Optional*. Set the common prefix of component names. If
             None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Omit the given attributes from the
               description.

:type omit: sequence


:returns: A string containing the description.

.. seealso:: :func:`cf.dump`        
'''
        return str(self)
    #--- End: def

    def evaluate(self, x):
        '''
Evaluate the comparison for a given object.

:parameter x: An object.

:type x: `object`

:returns: Typically a boolean or array of booleans. More
          generally, the output from the appropriate relation
          operators (``==, !=, <, >, <=, >=``) for the given
          variable, x.
    
**Examples:**

>>> c = cf.Comparison(lt = 5.5)
>>> c.evaluate(6)
False

>>> c = cf.Comparison(inside = (1,2))
>>> array = numpy.arange(4)
>>> c.evaluate(array)
array([False,  True,  True, False], dtype=bool)
'''
        if self.relation == 'lt':
            return x <  self.value
        elif self.relation == 'le':
            return x <= self.value
        elif self.relation == 'gt':
            return x >  self.value
        elif self.relation == 'ge':
            return x >= self.value
        elif self.relation == 'eq':
            return x == self.value
        elif self.relation == 'ne':
            return x != self.value
        elif self.relation == 'inside':
            if self.value[0] < self.value[1]:
                return (x >= self.value[0]) & (x <= self.value[1])
            else:
                return (x >= self.value[1]) & (x <= self.value[0])    
        elif self.relation == 'outside':
            if self.value[0] < self.value[1]:
                return (x <  self.value[0]) | (x >  self.value[1])
            else:
                return (x <  self.value[1]) | (x <  self.value[0])    
    #--- End: def

#--- End: class


def lt(value):
    '''
    
Create a :class:`~cf.Comparison` object for evaluating whether a
variable is strictly less than the given value.

:parameter value: The value which a variable is to be compared with.

:type value: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.lt(5)
>>> c
<CF Comparison: x lt 5>
>>> c.evaluate(4)
True
>>> c.evaluate(5)
False
    '''
    return Comparison(lt=value)
#--- End: def
    
def le(value):
    '''
    
Create a :class:`~cf.Comparison` object for evaluating whether a
variable is less than or equal to the given value.

:parameter value: The value which a variable is to be compared with.

:type value: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.le(5)
>>> c
<CF Comparison: x le 5>
>>> c.evaluate(5)
True
>>> c.evaluate(6)
False
'''
    return Comparison(le=value)
#--- End: def
    
def gt(value):
    '''
      
Create a :class:`~cf.Comparison` object for evaluating whether a
variable is greater than the given value.

:parameter value: The value which a variable is to be compared with.

:type value: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.gt(5)
>>> c
<CF Comparison: x gt 5>
>>> c.evaluate(6)
True
>>> c.evaluate(5)
False
'''
    return Comparison(gt=value)
#--- End: def
    
def ge(value):
    '''
     
Create a :class:`~cf.Comparison` object for evaluating whether a
variable is greater than or equal to the given value.

:parameter value: The value which a variable is to be compared with.

:type value: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.ge(5)
>>> c
<CF Comparison: x ge 5>
>>> c.evaluate(5)
True
>>> c.evaluate(4)
False
'''
    return Comparison(ge=value)
#--- End: def
    
def eq(value):
    '''
    
Create a :class:`~cf.Comparison` object for evaluating whether a variable
is equal to the given value.

:parameter value: The value which a variable is to be compared with.

:type value: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.eq(5)
>>> c
<CF Comparison: x eq 5>
>>> c.evaluate(5)
True
>>> c.evaluate(4)
False
'''
    return Comparison(eq=value)
#--- End: def
    
def ne(value):
    '''
    
Create a :class:`~cf.Comparison` object for evaluating whether a
variable is not equal to the given value.

:parameter value: The value which a variable is to be compared with.

:type value: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.ne(5)
>>> c
<CF Comparison: x ne 5>
>>> c.evaluate(4)
True
>>> c.evaluate(5)
False
'''
    return Comparison(ne=value)
#--- End: def
    
def inside(value0, value1):
    '''
    
Create a :class:`~cf.Comparison` object for evaluating whether a
variable is inside the given range.

:parameter value0: The lower bound of the range which a variable is to
                   be compared with.

:type value0: object

:parameter value1: The upper bound of the range which a variable is to
                   be compared with.

:type value1: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.inside(5,7)
>>> c
<CF Comparison: x inside (5, 7)>
>>> c.evaluate(6)
True
>>> c.evaluate(4)
False

'''

    return Comparison(inside=(value0, value1))
#--- End: def

def outside(value0, value1):
    '''
    
Create a :class:`~cf.Comparison` object for evaluating whether a
variable is outside the given range.

:parameter value0: The lower bound of the range which a variable is to
                   be compared with.

:type value0: object

:parameter value1: The upper bound of the range which a variable is to
                   be compared with.

:type value1: object

:returns: A :class:`~cf.Comparison` object which will evaluate whether
          or not the comparison evaluates to True.

**Examples:**

>>> c = cf.outside(5)
>>> c
<CF Comparison: x outside (5, 7)>
>>> c.evaluate(4)
True
>>> c.evaluate(6)
False
'''
    return Comparison(outside=(value0, value1))
#--- End: def

def RTOL(*new_default):
    '''

Return or set the default value of relative tolerance for testing
numerically tolerant equality.

**Examples:**

>>> cf.RTOL()
1.0000000000000001e-05
>>> cf.RTOL(1e-10)
>>> cf.RTOL()
1e-10

'''
    if new_default:
        _NUMERIC_TOLERANCE['RTOL'] = new_default[0]
    else:
        return _NUMERIC_TOLERANCE['RTOL']
#--- End: def

def ATOL(*new_default):
    '''

Return or set the default value of absolute tolerance for testing
numerically tolerant equality.

**Examples:**

>>> cf.ATOL()
1e-08
>>> cf.ATOL(1e-10)
>>> cf.ATOL()
1e-10
'''
    if new_default:
        _NUMERIC_TOLERANCE['ATOL'] = new_default[0]
    else:
        return _NUMERIC_TOLERANCE['ATOL']
#--- End: def


def equals(x, y, rtol=None, atol=None):
    '''

Ascertain if two objects are equal using numerically tolerant equality
where appropiate.

If the first argument (`x`) has an `equals` method, then it is
used. In this case ``equals(x,y)`` is equivalent to ``x.equals(y)``.

:argument x: The left hand side of the comparison.

:type x: object

:argument y: The right hand side of the comparison.
      
:type y: object
                           
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the absolute tolerance to this value for all such
               comparisons.

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the relative tolerance to this value for all such
               comparisons.

:returns: True or False.

**Examples:**

>>> x
<CF Field: rain(10,20)>
>>> cf.equals(x,x)
True

>>> cf.equals(1.0, 1.0)
True
>>> cf.equals(1.0, 33)
False

>>> cf.equals('a', 'a')
True
>>> cf.equals('a', 'b')
False

>>> type(x), x.dtype
(<type 'numpy.ndarray'>, dtype('int64'))
>>> y=x.copy()
>>> cf.equals(x, y)
True
>>> cf.equals(x, x+1)
False

>>> class A(object): pass
>>> a=A()
>>> b=A()
>>> cf.equals(a, a)
True
>>> cf.equals(a, b)
False

'''

    if rtol is None:
        rtol = RTOL()
    if atol is None:
        atol = ATOL()

    try:
        # x has a callable equals method
        x.equals
    except AttributeError:
        try:
            x.__iter__
        except AttributeError:
            try:
                y.__iter__
            except AttributeError: 
                try:
                    # x and y are both numbers
                    return abs(x-y) <= atol + rtol*abs(y)
                except TypeError:
                    # At least one of x and y is not a number
                    return x == y
            else:                        
                # y is a sequence
                return numpy.array_allclose(x, y, rtol=rtol, atol=atol)
            #--- End: try
        else:    
            # x is a sequence
            return numpy.array_allclose(x, y, rtol=rtol, atol=atol)
        #--- End: try
    else: 
        try:
            return x.equals(y, rtol=rtol, atol=atol)
        except TypeError:
            return x == y
    #--- End: try
#--- End: def

def dump(x, id=None, nc=False, omit=()):
    '''
    
Print the full description of an object using its callable `dump`
method.
    
This function is exactly equivalent to:

>>> print x.dump(...)

or, if the object has no `dump` method:

>>> print x

The parameters are only used if the object's `dump` method is being
utilized.

:parameter x: `Optional`. The object to print.

:type x: object

:parameter id: `Optional`. The same as the `id` parameter for the
               object's `dump` method.

:type id: str

:parameter omit: `Optional`. The same as the `omit` parameter for the
                 object's `dump` method.

:type omit: sequence


:returns: None

    '''
    try:
        x.dump
    except AttributeError:
        print str(x)
    else:
        print x.dump(id=id, omit=omit)
#--- End: def
        

def first_element(input):
    '''Find the first element of the input. If the input is a scalar,
    then the first element is that scalar.'''
    try:
        input.__iter__
    except:
        # Return a non-iterable scalar
        return input
    else:
        try:
            # Return the 1st element of a numpy.ndarray-like
            return input[(0,) * input.ndim]
        except AttributeError:
            try:
                # Return the 1st element of a list or tuple
                return input[0]
            except IndexError:
                # Return an empty list or tuple
                return input
#---End: def

def last_element(input):
    '''Find the last element of the input. If the input is a scalar,
    then the first element is that scalar.'''
    try:
        input.__iter__
    except:
        # Return a non-iterable scalar
        return input
    else:
        try:
            # Return the last element of a numpy.ndarray - like
            return input[(-1,) * input.ndim]
        except AttributeError:
            try:
                # Return the last element of a list or tuple
                return input[-1]
            except IndexError:
                # Return an empty list or tuple
                return input
#---End: def

def array_allclose(a1, a2, rtol=None, atol=None):
    '''

True if two arrays have the same shape and elements to within
numerical tolerance, False otherwise.

The tolerance values are positive, typically very small numbers. The
relative difference (`rtol` * abs(a2`)) and the absolute difference
`atol` are added together to compare against the absolute difference
between `a1` and `a2`.

Parameters
----------
a1, a2 : array_like
    Input arrays to compare.
rtol : float
    The relative tolerance parameter (see numpy.allclose).
atol : float
    The absolute tolerance parameter (see numpy.allclose).

Returns
-------
b : bool
    Returns True if the arrays are equal to within numerical
    tolerance.

Examples
--------
>>> array_allclose([1, 2], [1, 2])
True
>>> array_allclose(np.array([1, 2]), np.array([1, 2]))
True
>>> array_allclose([1, 2], [1, 2, 3])
False
>>> array_allclose([1, 2], [1, 4])
False
'''      
    try:
        a1 = numpy.asarray(a1)
        a2 = numpy.asarray(a2)
    except:
        return False

    if a1.shape != a2.shape:
        return False

    if rtol is None:
        rtol = RTOL()
    if atol is None:
        atol = ATOL()                                              

    try:
        return numpy.allclose(a1, a2, rtol=rtol, atol=atol)
    except IndexError:
        return numpy.all(a1 == a2)
#--- End: def

def _files(openfile):
    _open_files.append(openfile)
#--- End: def

def close():
    for f in _open_files:
        f.close()
    _open_files[:] = []
#--- End: def
